;; Now, suppose I want to track the object nesting level. Normally, this would
;; require having a specific visiting mechanism for alternatively incrementing
;; and decrementing the level each and every time we start/finish visiting an
;; object. But CLOS already has an infrastructure that let us do so very
;; simply.

(defgeneric map-object (func object)
  (:method (func (object standard-object))
    (funcall func  object)
    (mapc
	(lambda (slot)
	  (funcall #'map-object func
		   ;; That's where something should normally be done about the
		   ;; current nesting level.
		   ;; #### NOTE: SB-MOP is for SBCL.
		   (slot-value object (sb-mop:slot-definition-name slot))))
      (sb-mop:class-direct-slots (class-of object))))
  (:method (func (object list))
    (mapc func object)))


(defclass wheel () ())

(defclass body ()
  ((wheels :initform (loop :for i :from 1 :upto 4
			   :for wheel = (make-instance 'wheel)
			   :collect wheel)
	   :accessor wheels)))

(defclass engine () ())

(defclass porshe ()
  ((engine :initform (make-instance 'engine)
	   :accessor engine)
   (body :initform (make-instance 'body)
	 :accessor body)))


;; Here, I'm going to modify the visiting mechanism, indeed, but dynamically
;; and without having to touch the original one. CLOS lets you to dynamically
;; add/remove methods.
(defun main ()
  (let* ((porshe (make-instance 'porshe))
	 (depth 0)
	 ;; Just before calling a map-object primary method, let's increment
	 ;; the nesting level...
	 (before-method (defmethod map-object :before (func object)
			  (incf depth)))
	 ;; and just after calling a map-object primary method, let's
	 ;; decrement the nesting level.
	 (after-method (defmethod map-object :after (func object)
			 (decf depth))))
    ;; Note that I've just dynamically modified the behavior of my mapping
    ;; process.
    (map-object (lambda (obj)
		  (format t "~S, current level: ~A~%"
		    (class-name (class-of obj)) depth))
		porshe)
    ;; Let's remove this additionnal behavior and we're done.
    (remove-method #'map-object before-method)
    (remove-method #'map-object after-method))
  (values))
